/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CORE_OPERATOR_GENERICOPERATOR_H_
#define AIDGE_CORE_OPERATOR_GENERICOPERATOR_H_

#include <memory>
#include <vector>
#include <string>
#include <cassert>
#include <cstring>

#include "aidge/graph/Node.hpp"
#include "aidge/operator/OperatorTensor.hpp"
#include "aidge/utils/DynamicAttributes.hpp"
#include "aidge/utils/Registrar.hpp"
#include "aidge/utils/Types.h"


namespace Aidge {
class GenericOperator_Op
    : public OperatorTensor,
      public Registrable<GenericOperator_Op, std::string, std::unique_ptr<OperatorImpl>(std::shared_ptr<GenericOperator_Op>)>,
      public DynamicAttributes {
private:
    using ComputeDimsFunc = std::function<std::vector<std::vector<size_t>>(const std::vector<std::vector<size_t>>&)>;

    ComputeDimsFunc mComputeOutputDims;

public:
    GenericOperator_Op(const char *type, IOIndex_t nbData, IOIndex_t nbParam, IOIndex_t nbOut)
        : OperatorTensor(type, nbData, nbParam, nbOut)
    {}

    /**
     * @brief Copy-constructor. Copy the operator attributes and its output tensor(s), but not its input tensors (the new operator has no input associated).
     * @param op Operator to copy.
     */
    GenericOperator_Op(const GenericOperator_Op& op)
        : OperatorTensor(op)
    {}

    /**
     * @brief Clone the operator using its copy-constructor.
     * @see Operator::GenericOperator_Op
     */
    std::shared_ptr<Operator> clone() const override {
        return std::make_shared<GenericOperator_Op>(*this);
    }

    // Helper functions that can be used with setComputeOutputDims():
    static const ComputeDimsFunc Identity;

    void setComputeOutputDims(ComputeDimsFunc func) {
        mComputeOutputDims = func;
    }


    void computeOutputDims() override final {
        if (mComputeOutputDims) {
            std::vector<std::vector<size_t>> inputsDims(nbInputs(), std::vector<size_t>());
            for (std::size_t i = 0; i < nbInputs(); ++i) {
                if (getInput(i)) {
                    inputsDims[i] = getInput(i)->dims();
                }
            }

            const auto& outputsDims = mComputeOutputDims(inputsDims);
            assert(outputsDims.size() == nbOutputs() && "The provided ComputeDimsFunc function returns the wrong number of outputs");
            for (std::size_t i = 0; i < nbOutputs(); ++i) {
                mOutputs[i]->resize(outputsDims[i]);
            }
        }
        else {
            assert(false && "Cannot compute output dim of a GenericOperator");
        }
    }

    bool outputDimsForwarded() const override final {
        if (mComputeOutputDims) {
            return !(mOutputs[0]->empty());
        }
        else {
            assert(false && "GenericOperator cannot forward dims");
            return false;
        }
    }


    ~GenericOperator_Op() = default;

    void setBackend(const std::string & /*name*/) override { printf("setBackend: not available yet.\n"); }
    void setDataType(const DataType& /*datatype*/) const override { printf("setDataType: not available yet.\n"); }
    void forward() override final {
        if(mImpl){
            mImpl->forward();
        }else{
            printf("forward: No implementation is linked.\n");
        }
    }
    void backward() override final {
        if(mImpl){
            mImpl->backward();
        }else{
            printf("backward: No implementation is linked.\n");
        }
    }
};

/**
 * @brief Fictive custom operator not associated with any implementation.
 * Allows to import unknown operators and simulate new ones.
 * @param type Type of the fictive operator.
 * @param nbDataIn Number of input data.
 * @param nbIn Number input data + number of learnt parameters.
 * @param nbOut Number of output data.
 * @param name (optional) name of the Operator.
 * @return std::shared_ptr<Node> Node associated with the Generic Operator.
 */
inline std::shared_ptr<Node> GenericOperator(const char *type, IOIndex_t nbData, IOIndex_t nbParam, IOIndex_t nbOut,
                                             const std::string& name = "") {
    return std::make_shared<Node>(std::make_shared<GenericOperator_Op>(type, nbData, nbParam, nbOut), name);
}
}  // namespace Aidge

#endif /* AIDGE_CORE_OPERATOR_GENERICOPERATOR_H_ */
