/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CORE_OPERATOR_MATMUL_H_
#define AIDGE_CORE_OPERATOR_MATMUL_H_

#include <array>
#include <cmath>
#include <numeric>
#include <memory>
#include <vector>

#include "aidge/utils/Types.h"
#include "aidge/data/Tensor.hpp"
#include "aidge/graph/Node.hpp"
#include "aidge/operator/OperatorTensor.hpp"
#include "aidge/operator/Producer.hpp"
#include "aidge/utils/StaticAttributes.hpp"
#include "aidge/utils/Registrar.hpp"

namespace Aidge {
enum class MatMulAttr { OutChannels };

class MatMul_Op : public OperatorTensor,
              public Registrable<MatMul_Op,
                                 std::string,
                                 std::unique_ptr<OperatorImpl>(const MatMul_Op &)>,
              public StaticAttributes<MatMulAttr, DimSize_t> {
public:
    static constexpr const char* Type = "MatMul";

    MatMul_Op() = delete;

    using Attributes_ = StaticAttributes<MatMulAttr, DimSize_t>;
    template <MatMulAttr e> using attr = typename Attributes_::template attr<e>;

    MatMul_Op(DimSize_t out_channels)
            : OperatorTensor(Type, 1, 1, 1),
            Attributes_(
                attr<MatMulAttr::OutChannels>(out_channels))
    {}

    /**
     * @brief Copy-constructor. Copy the operator attributes and its output tensor(s), but not its input tensors (the new operator has no input associated).
     * @param op Operator to copy.
     */
    MatMul_Op(const MatMul_Op& op)
        : OperatorTensor(op),
          Attributes_(op)
    {
        mImpl = op.mImpl ? Registrar<MatMul_Op>::create(op.mOutputs[0]->getImpl()->backend())(*this) : nullptr;
    }

    /**
     * @brief Clone the operator using its copy-constructor.
     * @see Operator::MatMul_Op
     */
    std::shared_ptr<Operator> clone() const override {
        return std::make_shared<MatMul_Op>(*this);
    }


    void computeOutputDims() override final {
        bool associated = true;
        for (IOIndex_t i = 0; i < nbInputs(); ++i) {
            if (!getInput(i)) {
                AIDGE_THROW_OR_ABORT(std::runtime_error, "Every input should be associated with a Tensor");
            }
            associated &= !(getInput(i)->empty());
        }
        if (associated) {
            // <batch, OutChannels>
            mOutputs[0]->resize({getInput(0)->dims()[0], this->template getAttr<MatMulAttr::OutChannels>()});
        }
    }


    void setBackend(const std::string& name) override {
        mImpl = Registrar<MatMul_Op>::create(name)(*this);
        mOutputs[0]->setBackend(name);

        // FIXME: temporary workaround
        getInput(0)->setBackend(name);
        getInput(1)->setBackend(name);
    }

    static const std::vector<std::string> getInputsName(){
        return {"data_input", "weight"};
    }
    static const std::vector<std::string> getOutputsName(){
        return {"data_output"};
    }
};

inline std::shared_ptr<Node> MatMul(DimSize_t inChannels, DimSize_t outChannels, const std::string& name = "") {
    // FIXME: properly handle default w initialization in every cases
    auto matmul = std::make_shared<Node>(std::make_shared<MatMul_Op>(outChannels), name);
    addProducer(matmul, 1, std::array<DimSize_t, 2>({outChannels, inChannels}), "w");
    return matmul;
}
} // namespace Aidge

namespace {
template <>
const char *const EnumStrings<Aidge::MatMulAttr>::data[] = {"OutChannels"};
}

#endif /* AIDGE_CORE_OPERATOR__MATMUL_H_ */
